<?php

/*****************************************************************************

    Mysta - Tiny web framework.

    Copyright (C) 2012  Fabien LOISON

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/


/**
 * The router.
 */
class Router
{
	private $_routes = array();

	/**
	 * Register a new route into the router.
	 *
	 * @param $pattern
	 *   The route pattern (e.g. '/hello/<str:name>')
	 *   
	 *   Variables pattern:
	 *     - integer variable: <int:variable_name>
	 *     - string variable: <str:variable_name>
	 *     - path variable (like strings, but matches '/' too): <path:variable_name>
	 *
	 *   NOTE: the route pattern can contain regular expression.
	 *
	 * @param $callback
	 *   The callback function (called when the route match the pattern).
	 */
	public function add_route($pattern, $callback)
	{
		$this->_routes[] = array(
			'pattern' => $pattern,
			'regexp' => $this->_mk_regexp($pattern),
			'callback' => $callback);
		//"folders"
		if (preg_match('#^(.+)/$#', $pattern, $matches))
		{
			$this->_routes[] = array(
				'pattern' => $matches[1],
				'regexp' => $this->_mk_regexp($matches[1]),
				'callback' => '_folder_redirect');
		}
	}

	/**
	 * Call the function whose route pattern matches the given route.
	 *
	 * @param $route
	 *   The route (e.g. '/hello/world').
	 *
	 * @return
	 *   The content of the page.
	 */
	public function route($route)
	{
		foreach ($this->_routes as $route_pattern)
		{
			if (preg_match($route_pattern['regexp'], $route, $matches))
			{
				$args = array();
				foreach ($matches as $key => $value)
				{
					if (preg_match('/^[A-Z][A-Z0-9_]*$/i', $key))
					{
						$args[] = $value;
					}
				}
				return call_user_func_array($route_pattern['callback'], $args);
			}
		}
		//No route... 404
		abort(404);
	}

	/**
	 * Generate the URL for the wanted route.
	 *
	 * @param $route_name
	 *   The route name (the name of the callback function)
	 *
	 * @param $args
	 *   Associative array with all the needed arguments of the route.
	 *   (e.g. array('name' => 'world'))
	 *
	 * @return
	 *   The URL.
	 */
	public function url_for($route_name, $args=array())
	{
		foreach ($this->_routes as $route)
		{
			if ($route['callback'] == $route_name)
			{
				$url = $route['pattern'];
				foreach ($args as $key => $value)
				{
					$url = preg_replace('/<(int|str|path):' . $key . '>/iU', $value, $url);
				}
				if (get_current_route() != '')
				{
					return get_base_uri() . $url;
				}
				else
				{
					return $_SERVER['REQUEST_URI'] . '?' . $url;
				}
			}
		}
		//No route... 500
		abort(500);
	}

	/**
	 * Make a regular expression from the given route pattern.
	 *
	 * @param $pattern
	 *   The route pattern.
	 *
	 * @return
	 *   The regular expression.
	 */
	private function _mk_regexp($pattern)
	{
		$pattern = preg_replace('/<int:([A-Z_][A-Z0-9_]*)>/iU', '(?P<$1>\d+)', $pattern);
		$pattern = preg_replace('/<str:([A-Z_][A-Z0-9_]*)>/iU', '(?P<$1>[A-Za-z0-9_. +-]+)', $pattern);
		$pattern = preg_replace('/<path:([A-Z_][A-Z0-9_]*)>/iU', '(?P<$1>[A-Za-z0-9_. +/-]+)', $pattern);
		return '#^' . $pattern . '$#';
	}
}


/**
 * Generate the URL for the wanted route.
 *
 * @param $route_name
 *   The route name (the name of the callback function)
 *
 * @param $args
 *   Associative array with all the needed arguments of the route.
 *   (e.g. array('name' => 'world'))
 *
 * @return
 *   The URL.
 */
function url_for($route_name, $args=array())
{
	global $router;
	return $router->url_for($route_name, $args);
}


/**
 * Callback for static content.
 *
 * NOTE: this function will render the page immediatly.
 *
 * @param $filename
 *   The name of the static file.
 */
function staticfile($filename, $prefix='./static/')
{
	if (preg_match('#(\.\./|\.php|/\.ht)#s', $filename))
	{
		abort(404);
	}

	$filepath = $prefix . $filename;
	if (is_file($filepath))
	{
		header('Content-Length: ' . filesize($filepath));
		header('Content-Type: ' . give_me_the_mime($filepath));
		readfile($filepath);
		die();
	}
	else
	{
		abort(404);
	}
}


/**
 * Callback function for redirecting "folder" routes.
 *
 * e.g. redirect '/hello' to '/hello/'
 */
function _folder_redirect()
{
	redirect($_SERVER['REQUEST_URI'] . '/');
}


//
$router = new Router();
$router->add_route('/static/<path:filename>', 'staticfile');
